/*
 This file is part of the LifeV library
 Copyright (C) 2001,2002,2003,2004 EPFL, INRIA and Politechnico di Milano

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef _NONLINRICHARDSON_HPP
#define _NONLINRICHARDSON_HPP

#include <algorithm> // for min and max
#include <life/lifealg/linesearch_parabolic.hpp>
#include <life/lifealg/linesearch_cubic.hpp>
#include <life/lifealg/generalizedAitken.hpp>

namespace LifeV
{
template <class Fct, class VectorType, class Real, class Norm>
int nonLinRichardson( VectorType& sol,
                      Fct& f,
                      const Norm& norm,
                      Real abstol,
                      Real reltol,
                      int& maxit,
                      Real eta_max,
                      int linesearch,
                      std::ofstream& out_res,
                      const Real& time )
//                      const Real omega )
{
    /*
      sol            :  the solution
      maxit          :  input: maximum iterations, output: nb of iterations
      abstol, reltol :  the stoping criteria is abstol+reltol*norm(residual_0),
      eta_max        :  Maximum error tolerance for residual in linear solver.

      The linear solver terminates when the relative
      linear residual is smaller than eta*| f(sol) |.

      The value linear_rel_tol send for the relative tolerance
      to the linear solver is therefore eta. eta is determined
      by the modified Eisenstat-Walker formula if etamax > 0.

      linesearch     :  for now consider only the case linesearch=0
      (coded but not theoretically analysed)

      omega          :  default relaxation parameter to be passed to Aitken.
      if omega is negative, then its absolute value is
      taken as constant relaxation parameter
    */

    /*
      max_increase_res: maximum number of successive increases in residual
      before failure is reported
    */

    const int max_increase_res = 5;

    /*
      Parameters for the linear solver, gamma: Default value = 0.9
    */

    const Real gamma   = 0.9;

    //----------------------------------------------------------------------


    int iter = 0;

    int nDofFS = sol.size();

    VectorType residual ( sol.size() );
    VectorType step     ( sol.size() );

    step = ZeroVector( step.size() );

    Real normResOld = 1;

    std::cout << "------------------------------------------------------------------" << std::endl;
    std::cout << "  NonLinRichardson: starting " << std::endl;
    std::cout << "------------------------------------------------------------------" << std::endl;
    f.evalResidual( residual, sol, iter );

    Real normRes      = norm( residual );
    Real stop_tol     = abstol + reltol*normRes;
    Real linearRelTol = fabs(eta_max);
    Real eta_old;
    Real eta_new;
    Real ratio;
    Real slope;
    Real linres;
    Real lambda;

    //

    out_res << std::scientific;
    out_res << "# time = ";
    out_res << time << "   " << "initial norm_res " <<  normRes
            << " stop tol = " << stop_tol
            << "initial norm_sol "
            << norm_2(sol) << std::endl;
    out_res << "#iter      disp_norm       step_norm       residual_norm" << std::endl;

    while ( normRes > stop_tol && iter < maxit )
    {
        std::cout << std::endl;
        std::cout << "------------------------------------------------------------------" << std::endl;
        std::cout << "  NonLinRichardson: iter = " << iter
                  << ", residual = " << normRes
                  << ", stoping tolerance = " << stop_tol << std::endl;
        std::cout << "------------------------------------------------------------------" << std::endl;
        std::cout << std::endl;

        iter++;

        ratio = normRes/normResOld;
        normResOld = normRes;
        normRes = norm(residual);

        f.solveJac(step, -1.*residual, linearRelTol);

        linres = linearRelTol;

        lambda = 1.;
        slope  = normRes * normRes * ( linres * linres - 1 );

        switch ( linesearch )
        {
            case 0: // no linesearch
                sol += step;
                f.evalResidual( residual, sol, iter);
                normRes = norm( residual );
                break;
            case 1:
                lineSearch_parab( f, norm, residual, sol, step, normRes, lambda, iter );
                break;
            case 2:  // recommended
                lineSearch_cubic( f, norm, residual, sol, step, normRes, lambda, slope, iter );
                break;
            default:
                std::cout << "Unknown linesearch \n";
                exit( 1 );
        }


        out_res   << std::setw(5) << iter
                  << std::setw(15) << norm_2  (sol)
                  << std::setw(15) << norm_2  (step);

        f.displacementOnInterface();
        normRes = norm( residual );

        out_res << std::setw(15) << normRes << std::endl;

        if ( eta_max > 0 )
        {
            eta_old = linearRelTol;
            eta_new = gamma * ratio * ratio;
            if ( gamma * eta_old * eta_old > .1 )
            {
                eta_new = std::max<Real>( eta_new, gamma * eta_old * eta_old );
            }
            linearRelTol = std::min<Real>( eta_new, eta_max );
            linearRelTol = std::min<Real>( eta_max,
                                             std::max<Real>( linearRelTol,
                                                             .5 * stop_tol / normRes ) );
            std::cout << "    Newton: forcing term eta = " << linearRelTol << std::endl;
        }
    }

    if ( normRes > stop_tol )
    {
        std::cout << "!!! NonLinRichardson: convergence fails" << std::endl;
        maxit = iter;
        return 1;
    }

    f.displacementOnInterface();
    std::cout << "------------------------------------------------------------------" << std::endl;
    std::cout << "--- NonLinRichardson: convergence (" << normRes
              <<") in " << iter << " iterations\n\n";
    std::cout << "------------------------------------------------------------------" << std::endl;

    maxit = iter;

    return 0;
}
}
#endif
