/* -*- mode: c++ -*-

  This file is part of the LifeV library

  Author(s): Christophe Prud'homme <christophe.prudhomme@epfl.ch>
       Date: 2004-11-18

  Copyright (C) 2004 EPFL

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/**
   \file FSISolver.hpp
   \author Christophe Prud'homme <christophe.prudhomme@epfl.ch>
   \date 2004-11-18
 */
#ifndef __FSISolver_H
#define __FSISolver_H 1

#include <life/lifearray/tab.hpp>
#include <life/lifesolver/FSIOperator.hpp>
#include <life/lifealg/nonLinRichardson.hpp>
#include <life/lifealg/newton.hpp>

namespace LifeV
{
/*!
  \class FSISolver
  \brief solver for Fluid-Structure Interaction

  \c FSISolver uses the FSI operators whose base class is \c FSIOperator to
  solve the FSI problem.  It behaves from an interface point of view
  very much like a NS solver or solid solver.

  The temporal loop is externalized: the member function \c
  iterate(time) must be called at each time steps.  The time data is
  the one associated with the fluid.

  \c FSISolver allows to change the FSI operator using the \c setFSIOperator(name)
  member function that needs the \c name of the new FSI operator to be used.

  \todo Generic fluid and Structure solvers
  \todo Allow delayed initialization

  @author Christophe Prud'homme
  @see FSIOperator
*/
class FSISolver
{
public:

    /** @name Typedefs
     */
    //@{
//    typedef FSIOperator::fluid_type::value_type::source_type
    typedef FSIOperator::fluid_type::value_type::source_type fluid_source_type;
    typedef FSIOperator::fluid_type::value_type::source_type solid_source_type;
    typedef FSIOperator::fluid_bchandler_type fluid_bchandler_type;
    typedef FSIOperator::solid_bchandler_type solid_bchandler_type;
    //@}

    /** @name Constructors, destructor
     */
    //@{

    /*!
      \brief default/only constructor for the FSI solver

      \attention the last argument of the constructor gives a
      possibility to override the operator that is given by the \c
      data_file. An empty string (default) means that it uses the
      operator defined in \c data_file
      \todo allow to change the FSI operator on the fly
     */
    FSISolver( GetPot const& datafile,
               fluid_bchandler_type __bcu,
               solid_bchandler_type __bcd,
               fluid_bchandler_type __bchext,
               std::string __oper = "" );

    FSISolver( GetPot const& datafile,
               const DataNavierStokes<RegionMesh3D_ALE<LinearTetra> >& dataNavierStokes,
               fluid_bchandler_type __bcu,
               solid_bchandler_type __bcd,
               fluid_bchandler_type __bchext,
               std::string __oper = "" );


    FSISolver( GetPot const& datafile,
               std::string __oper = "" );

    //@}

    /** @name Operator overloads
     */
    //@{


    //@}


    //@}

    /** @name Accessors
     */
    //@{

    //! get the time step
    Real timeStep() const { return M_fluid->dt(); }

    //! get the final time
    Real timeEnd() const { return M_fluid->endtime(); }

    //! get the FSI operator
    oper_fsi_ptr const& operFSI() const { return M_oper; }

    //! get the displacement
    Vector const& displacement() const { return M_disp; }

    //! get access to the \c bchandler_type for the velocity
    fluid_bchandler_type& bcHandlerU() { return M_BCh_u; }

    //! get access to the \c bchandler_type for the velocity
    fluid_bchandler_type const& bcHandlerU() const { return M_BCh_u; }

    //! get access to the \c bchandler_type for the displacement
    solid_bchandler_type const& bcHandlerD() const { return M_BCh_d; }

    //@}

    /** @name  Mutators
     */
    //@{

    void setSourceTerms( fluid_source_type const& __f, solid_source_type const& __s )
        {
            M_fluid->setSourceTerm( __f );
            M_solid->setSourceTerm( __s );
        }

    /*!
      \brief set the FSI operator to be used to couple the fluid and the structure

      The name of the operator has to be passed as an argument

      \param __op FSI operator name
    */
    void setFSIOperator( std::string const& __op );
    oper_fsi_ptr FSIOper(){ return M_oper; }
    //@}

    void setFluidBC             (fluid_bchandler_type &bc_fluid);
    void setLinFluidBC          (fluid_bchandler_type &bc_dfluid);
    void setInvLinFluidBC       (fluid_bchandler_type &bc_dfluid_inv);
    void setHarmonicExtensionBC (fluid_bchandler_type &bc_he);
    void setSolidBC             (solid_bchandler_type &bc_solid);
    void setLinSolidBC          (solid_bchandler_type &bc_dsolid);
    void setInvLinSolidBC       (solid_bchandler_type &bc_dsolid_inv);
    void setReducedLinFluidBC   (fluid_bchandler_type &bc_dredfluid);
    void setInvReducedLinFluidBC(fluid_bchandler_type &bc_dredfluid_inv);

    /** @name  Methods
     */
    //@{

    void initialize( FSIOperator::fluid_type::value_type::Function const& __u0,
                     FSIOperator::solid_type::value_type::Function const& __d0,
                     FSIOperator::solid_type::value_type::Function const& __w0 )
        {
            M_oper->fluid().initialize(__u0);
            M_oper->solid().initialize(__d0,__w0);
        }

    void initialize( std::string velFName,
                     std::string pressName,
                     std::string velwName,
                     std::string depName,
                     std::string velSName,
                     double      Tstart = 0.)
        {
            M_oper->fluid().initialize(velFName, pressName, velwName, Tstart);
            M_oper->solid().initialize(depName, velSName, Tstart);
        }

    void iterate( Real time );

    void showMe()
        {
            M_fluid->showMe();
            M_solid->showMe();
        }

    //@}

private:

    //! forbid default constructor
    FSISolver();

    //! forbid copy constructor
    FSISolver( FSISolver const& );

    //! forbid copy operator
    FSISolver& operator=( FSISolver const& );

private:

    // be careful here: the BCs must be constructed before the solvers
    fluid_bchandler_type     M_BCh_u;
    fluid_bchandler_type     M_BCh_d;
    fluid_bchandler_type     M_BCh_mesh;

    fluid_bchandler_type     M_BCh_du;
    fluid_bchandler_type     M_BCh_dz;

    oper_fsi_ptr       M_oper;

    FSIOperator::fluid_type M_fluid;
    FSIOperator::solid_type M_solid;


    Vector             M_disp;
    Vector             M_velo;

    /* data */
    bool               M_firstIter;
    std::string        M_method;
    UInt               M_maxpf;
    Real               M_defomega;

    Real               M_abstol;
    Real               M_reltol;
    Real               M_etamax;

    int                M_linesearch;

    /* streams */
    std::ofstream out_iter;
    std::ofstream out_res;

};

}
#endif /* __FSISolver_H */
